classdef CDOASIS < handle
    %CDOASIS This class provides an interface to DOASIS spectral processing
    %functions. 
    %   Usage: First set the public class properties, then call Go to start     
    %   the processing. Use DisplayFitResult to get a result graphic. 
    
    properties
        DOASISpath % Path to the DOASIS install directory containing DoasConsole.exe.
        InterfaceDirectory % The directory which contains files needed for the DOASIS interface (e.g. empty fit scenarios) and where temporary files will be written.
        ScriptFilename % Filename (not including path) of the script that will be generated and passed to DoasConsole.exe
        OutputBuffer % Filename (not including path) of a temporary buffer for transfering output from DOASIS to the CDOASIS class (when QuickFit is off).
        TargetSpectrum % This is the spectrum or spectra to be processed. This can be (1) a single filename or (2) a cell array of filenames or (3) a CSpectrum object.
        CorrectOffset % Boolean indicating whether to include offset correction.
        OffsetSpectrumFilename % Filename of the offset spectrum
        CorrectDarkCurrent % Boolean indicating whether to include dark current correction.
        DarkCurrentSpectrumFilename % Filename of the Dark Current spectrum
        CorrectStrayLight % Boolean indicating whether to subtract a stray light baseline.
        StrayLightPixelRange % 2 element array indicating the pixel range in which the stray light level will be determined.
        TakeLogarithm % Boolean indicating whether or not to take the logarithm of the target spectrum.
        SpectraAreAveraged % Boolean indicating whether spectra are integrated by averaging rather than summing. This applies to the targets and all references as well.
        RemoveOffsetFromDarkCurrent % Boolean indicating whether to remove the electronic offset from the dark current spectrum.
        SavePrefitTarget % Boolean indicating whether to save each target spectrum prior to running the fit for manual analysis later on.
        PrefitDirectory % Directory in which to save the prefit targets if desired.
        DoFit % Boolean indicating whether to perform a DOAS fit
        FitScenario % CFitScenario object containing the fit scenario to use in the fit.        
        QuickFit % Boolean indicating whether to do a 'QuickFit'. Use this option to speed up processing by 3x, but note that only FitCoefficients, FitCoefficientErrors, Shifts, Squeezes, and Chi2 will be available.        
    end
    
    properties (SetAccess = private)
        NTargets % The number of valid target spectra that were processed.
        FitCoefficients % Vector of fit coefficient results (from last fit). If more than one target is specified, returns a matrix in which each column corresponds to a different target.
        FitCoefficientErrors % Vector of fit coefficient errors (from last fit). If more than one target is specified, returns a matrix in which each column corresponds to a different target.
        Chi2 % Chi Squared (from last fit). If more than one target is specified, returns a horizontal array in which each element corresponds to a different target.
        Shifts % Vector of shift values applied to the cross-sections to obtain the best fit (from last fit). If more than one target is specified, returns a matrix in which each column corresponds to a different target.
        Squeezes % Vector of squeeze values applied to the cross-sections to obtain the best fit (from last fit). If more than one target is specified, returns a matrix in which each column corresponds to a different target.        
        FitPolynomial % Vector of polynomial coefficients used to obtain the best fit (from last fit). The vector has N+1 rows, where N is the polynomial order. If more than one target is specified, returns a matrix in which each column corresponds to a different target.
        PreFitTarget % Vector representing the TargetSpectrum after any modifications in preparation for fitting. Each row is a channel. If more than one target is specified, returns a matrix in which each column corresponds to a different target.
        FitResidual % Vector representing the FitResidual spectrum. Each row is a channel. If more than one target is specified, returns a matrix in which each column corresponds to a different target.
        FitCrossSections % Matrix holding fit results of the cross-sections. Each row is a channel. Each column is a cross-section. If more than one target is specified, the 3rd dimension holds results from subsequent targets.
    end
    
    methods
        function DOASIS = CDOASIS()
            %CDOASIS Construct an instance of this class            
            if exist('C:\Program Files (x86)\DOASIS\DoasConsole.exe', 'file') > 0
                DOASIS.DOASISpath = 'C:\Program Files (x86)\DOASIS\DoasConsole.exe';
            elseif exist('C:\Program Files\DOASIS\DoasConsole.exe', 'file') > 0
                DOASIS.DOASISpath = 'C:\Program Files\DOASIS\DoasConsole.exe';
            else
                disp('Warning: DOASIS was not found in the default installation directory. Please manually set the DOASISpath parameter.');
            end
            DOASIS.CorrectOffset = 1;
            DOASIS.CorrectDarkCurrent = 1;
            DOASIS.DoFit = 1;
            DOASIS.SpectraAreAveraged = 0;
            DOASIS.RemoveOffsetFromDarkCurrent = 0;
            DOASIS.TargetSpectrum = [];
            DOASIS.QuickFit = 0;
            DOASIS.SavePrefitTarget = 0;
            DOASIS.ScriptFilename = 'dfit.js';
            DOASIS.OutputBuffer = 'dout.dat';
            DOASIS.InterfaceDirectory = [pwd filesep];
        end
        
        
        function BuildScript(DOASIS, ScriptFilename)
            % This function is called by DOASIS.Go to build the execution
            % script before then running it. However, it can also be called
            % from outside the Go function to build the script but not
            % execute it (useful for running manually in DOASIS).
            
            % Open the jscript file for writing           
            fid = fopen(ScriptFilename, 'wt');

            % import the namespaces
            fprintf(fid, 'import System; import DoasCore; import DoasCore.Spectra; import DoasCore.Script; import DoasCore.Math;\n\n');            

            % create a ISpectrum objects to hold the various spectra:
            fprintf(fid, 'var TargetSpec : ISpectrum = Specbar.GetSpectrum("Target");\n');
            fprintf(fid, 'var OffsetSpec : ISpectrum = Specbar.GetSpectrum("Offset");\n');
            fprintf(fid, 'var DarkCurrentSpec : ISpectrum = Specbar.GetSpectrum("DarkCurrent");\n'); 
            fprintf(fid, 'var ResultSpec : ISpectrum = Specbar.GetSpectrum("Result");\n');                      
            fprintf(fid, 'var ResidualSpec : ISpectrum = Specbar.GetSpectrum("Fit Residual");\n');
            
            for i = 1:length(DOASIS.FitScenario.CrossSections)
                fprintf(fid, ['var CSSpec' num2str(i) ' : ISpectrum = Specbar.GetSpectrum("Fit ' DOASIS.FitScenario.CrossSections{i}.Spectrum.Name '");\n']);
            end

            % tell the jscript how many target spectra to expect:      
            % if TargetSpectrum is a single filename, use that:
            if isstring(DOASIS.TargetSpectrum) || ischar(DOASIS.TargetSpectrum)
                if ~exist(DOASIS.TargetSpectrum, 'file')
                    disp(['ERROR: The target spectrum "' DOASIS.TargetSpectrum '" does not exist!']);
                    fprintf(fid, 'var Targets = 0;\n');
                    DOASIS.NTargets = 0;
                else
                    fprintf(fid, 'var Targets = 1;\n');
                    DOASIS.NTargets = 1;
                end
            elseif iscell(DOASIS.TargetSpectrum) %TargetSpectrum is a cell array of file names          
                DOASIS.NTargets = 0;
                for i = length(DOASIS.TargetSpectrum):-1:1
                    if ~exist(DOASIS.TargetSpectrum{i}, 'file')
                        disp(['Warning: The target spectrum "' DOASIS.TargetSpectrum{i} '" does not exist. Will try to load others in the cell array.']);
                        DOASIS.TargetSpectrum(i) = [];
                    else
                        DOASIS.NTargets = DOASIS.NTargets + 1;
                    end                                                             
                end
                fprintf(fid, ['var Targets = ' num2str(DOASIS.NTargets) ';\n']);
                sbfr = 'var TargetSpecFNs = [';                
                for i = 1:DOASIS.NTargets-1
                    fn = DOASIS.TargetSpectrum{i};
                    fn = strrep(fn, '\', '\\\\');
                    sbfr = [sbfr '"' fn '", '];
                end
                fn = DOASIS.TargetSpectrum{DOASIS.NTargets};
                fn = strrep(fn, '\', '\\\\');
                sbfr = [sbfr '"' fn '"];\n'];
                fprintf(fid, sbfr);
            else % The target spectrum is a CSpectrum, so save it first:
                DOASIS.TargetSpectrum.Save('target.std');                
                fprintf(fid, 'var Targets = 1;\n');
                DOASIS.NTargets = 1;
            end          
            
            % Load the offset and dark current spectra if needed
            if DOASIS.CorrectOffset || DOASIS.RemoveOffsetFromDarkCurrent
                osfn = DOASIS.OffsetSpectrumFilename;
                osfn = strrep(osfn, '\', '\\\\');
                fprintf(fid, ['OffsetSpec.Open("' osfn '");\n']);
                if DOASIS.SpectraAreAveraged
                    fprintf(fid, 'OffsetSpec = SpecMath.MulScalar(OffsetSpec, OffsetSpec.NumScans);\n');
                end                    
            end            
            if DOASIS.CorrectDarkCurrent
                dcfn = DOASIS.DarkCurrentSpectrumFilename;
                dcfn = strrep(dcfn, '\', '\\\\');
                fprintf(fid, ['DarkCurrentSpec.Open("' dcfn '");\n']);
                if DOASIS.SpectraAreAveraged
                    fprintf(fid, 'DarkCurrentSpec = SpecMath.MulScalar(DarkCurrentSpec, DarkCurrentSpec.NumScans);\n');
                end
                if DOASIS.RemoveOffsetFromDarkCurrent
                    fprintf(fid, 'DarkCurrentSpec = SpecMath.CorrectOffset(DarkCurrentSpec, OffsetSpec);\n');
                end
            end
            
            % open the fit scenario
            if isempty(DOASIS.FitScenario.FileName)
                DOASIS.FitScenario.BuildDOASISFitScenario('fitscen.fs');
            end
            fsfn = DOASIS.FitScenario.FileName;
            fsfn = strrep(fsfn, '\', '\\\\');
            fprintf(fid, ['var fit = DoasFit.Open("' fsfn '");\n']);                      
            
            % prepare the file system object for writing the output file:
            fprintf(fid, 'var fso = new ActiveXObject("Scripting.FileSystemObject");\n');   
            
            % if Prefits will be saved, make sure the directory exists
            if DOASIS.SavePrefitTarget
                if isempty(DOASIS.PrefitDirectory)
                    DOASIS.PrefitDirectory = [pwd filesep];                
                elseif DOASIS.PrefitDirectory(length(DOASIS.PrefitDirectory)) ~= filesep
                    DOASIS.PrefitDirectory = [DOASIS.PrefitDirectory filesep];
                end
                if ~exist(DOASIS.PrefitDirectory, 'dir')
                    mkdir(DOASIS.PrefitDirectory);
                end
            end
            
            % BEGIN THE PROCESSING LOOP ---------------------------------
            fprintf(fid, 'var target = 0;\n\n');
            fprintf(fid, 'while(target < Targets) {\n');                        
            
            % create a string to hold the output buffer:
            % this needs to be done in the loop or the bfr can overflow.
            fprintf(fid, 'var s = "";\n'); 
            
            % Open the next target spectrum:
            if isstring(DOASIS.TargetSpectrum) || ischar(DOASIS.TargetSpectrum)
                tsfn = DOASIS.TargetSpectrum;
                tsfn = strrep(tsfn, '\', '\\\\');
                fprintf(fid, ['TargetSpec.Open("' tsfn '");\n']);
            elseif iscell(DOASIS.TargetSpectrum) %TargetSpectrum is a cell array of file names
                fprintf(fid, 'TargetSpec.Open(TargetSpecFNs[target]);\n');
            else %TargetSpec is a CSpectrum object
                fprintf(fid, 'TargetSpec.Open(target.std);\n');                
            end            
            
            % multiply by numscans if needed:
            if DOASIS.SpectraAreAveraged
                fprintf(fid, 'TargetSpec = SpecMath.MulScalar(TargetSpec, TargetSpec.NumScans);\n');
            end
            
            % perform the offset and dark current corrections if needed
            fprintf(fid, 'TargetSpec = SpecMath.CorrectOffset(TargetSpec, OffsetSpec);\n');
            fprintf(fid, 'TargetSpec = SpecMath.CorrectDarkCurrent(TargetSpec, DarkCurrentSpec);\n');
            
            % Stray light correction
            if DOASIS.CorrectStrayLight                
                fprintf(fid, ['TargetSpec.MinChannel = ' num2str(DOASIS.StrayLightPixelRange(1)) ';\n']);
                fprintf(fid, ['TargetSpec.MaxChannel = ' num2str(DOASIS.StrayLightPixelRange(2)) ';\n']);
                fprintf(fid, 'var sl = TargetSpec.Average;\n');
                fprintf(fid, 'TargetSpec.MinChannel = 0;\n');
                fprintf(fid, 'TargetSpec.MaxChannel = TargetSpec.NChannel - 1;\n');
                fprintf(fid, 'TargetSpec = SpecMath.SubScalar(TargetSpec, sl);\n');
            end
            
            % Take the log
            if DOASIS.TakeLogarithm
                fprintf(fid, 'TargetSpec = SpecMath.Log(TargetSpec);\n');
            end
            
            % If desired, save the modified TargetSpec:
            if DOASIS.SavePrefitTarget
                if isstring(DOASIS.TargetSpectrum) || ischar(DOASIS.TargetSpectrum)
                    [~, fn, ext] = fileparts(DOASIS.TargetSpectrum);
                    pffn = [DOASIS.PrefitDirectory 'prefit_' fn ext];
                    pffn = strrep(pffn, '\', '\\\\');
                    fprintf(fid, ['TargetSpec.SaveAs("' pffn '");\n']);
                elseif iscell(DOASIS.TargetSpectrum)
                    dn = DOASIS.PrefitDirectory;
                    dn = strrep(dn, '\', '\\\\');
                    fprintf(fid, 'var n = TargetSpecFNs[target].lastIndexOf("\\\\");\n');
                    fprintf(fid, 'var fn: String;\n');                    
                    fprintf(fid, 'var dn: String;\n');
                    fprintf(fid, 'var pf: String;\n');                    
                    fprintf(fid, 'fn = TargetSpecFNs[target].substring(n+1, TargetSpecFNs[target].length);\n');
                    fprintf(fid, ['dn = "' dn '";\n']);
                    fprintf(fid, 'pf = "prefit_";\n');                    
                    fprintf(fid, 'dn = dn.concat(pf);\n');
                    fprintf(fid, 'dn = dn.concat(fn);\n');                                                          
                    fprintf(fid, 'TargetSpec.SaveAs(dn);\n');
                else %TargetSpec is a CSpectrum object
                    % try to derive a valid SpectrumID if none exists
                    if isempty(DOASIS.TargetSpec.SpectrumID)
                        DOASIS.TargetSpec.SpectrumID = AssignSpectrumID(DOASIS.TargetSpec.FileName);
                    end
                    if isempty(TargetSpec.SpectrumID)                    
                        dv = datevec(now);
                        eds = [num2str(dv(1), '%04d') '-' num2str(dv(2), '%02d') '-' num2str(dv(3), '%02d') 'T' num2str(dv(4), '%02d') '-' num2str(dv(5), '%02d') '-' num2str(round(dv(6)), '%02d')];                                
                        fn = [DOASIS.PrefitDirectory 'prefit_' eds '.std'];
                    else
                        fn = [DOASIS.PrefitDirectory 'prefit_' num2str(TargetSpec.SpectrumID) '.std'];
                    end
                    fn = strrep(fn, '\', '\\\\');
                    fprintf(fid, ['TargetSpec.SaveAs("' fn '");\n']);
                end
            end
            
            % increment the target counter
            fprintf(fid, 'target = target + 1;\n');
                                               
            % do the fit
            fprintf(fid, 'if(fit.DoFit(TargetSpec))\n{\n\n');
            
            % prepare the results if needed:
            if DOASIS.QuickFit == 0
                fprintf(fid, 'fit.PrepareFitResultSpectrum(ResultSpec);\n\n');           
            end                        
            
            % get access to the ReferencesInfo object
            fprintf(fid, 'var RefInfos = fit.ReferencesInfo;\n');            
            fprintf(fid, 'var d : double;\n');
            
            if DOASIS.QuickFit == 0
                % output the fit coefficients                        
                fprintf(fid, 's = s + "Fit Coefficients:\\r\\n";\n');
                for i = 1:length(DOASIS.FitScenario.CrossSections)
                    fprintf(fid, ['d = RefInfos[' num2str(i-1) '].GetCoefficientValue(0);\n']);                
                    fprintf(fid, 's = s + d.ToString() + "\\r\\n";\n');                
                end                        

                % output the fit errors
                fprintf(fid, 's = s + "Fit Coefficient Errors:\\r\\n";\n');
                for i = 1:length(DOASIS.FitScenario.CrossSections)
                    fprintf(fid, ['d = RefInfos[' num2str(i-1) '].GetCoefficientError(0);\n']);
                    fprintf(fid, 's = s + d.ToString() + "\\r\\n";\n');
                end

                % output the Chi Square
                fprintf(fid, 's = s + "ChiSquare:\\r\\n";\n');
                fprintf(fid, 's = s + (fit.ChiSquare).ToString() + "\\r\\n";\n');
            
                % output the shifts
                fprintf(fid, 's = s + "Shifts:\\r\\n";\n');
                for i = 1:length(DOASIS.FitScenario.CrossSections)
                    fprintf(fid, ['d = RefInfos[' num2str(i-1) '].GetCoefficientValue(1);\n']);
                    fprintf(fid, 's = s + d.ToString() + "\\r\\n";\n');
                end

                % output the squeezes
                fprintf(fid, 's = s + "Squeezes:\\r\\n";\n');
                for i = 1:length(DOASIS.FitScenario.CrossSections)
                    fprintf(fid, ['d = RefInfos[' num2str(i-1) '].GetCoefficientValue(2);\n']);
                    fprintf(fid, 's = s + d.ToString() + "\\r\\n";\n');
                end

                % output the polynomial:
                fprintf(fid, 's = s + "Polynomial:\\r\\n";\n');
                for i = 0:DOASIS.FitScenario.PolynomialOrder                               
                    fprintf(fid, ['s = s + fit.Polynomial.GetCoefficient(' num2str(i) ').ToString() + "\\r\\n";\n']);
                end

                % output the prefit target:
                fprintf(fid, 's = s + "PreFitTarget:\\r\\n";\n');
                for i = DOASIS.FitScenario.ChannelRangeLow+1:DOASIS.FitScenario.ChannelRangeHigh+1
                    fprintf(fid, ['s = s + TargetSpec.Intensity[' num2str(i) '].ToString() + "\\r\\n";\n']);
                end

                % output the fit residual spectrum:
                fprintf(fid, 's = s + "Residual:\\r\\n";\n');
                for i = DOASIS.FitScenario.ChannelRangeLow+1:DOASIS.FitScenario.ChannelRangeHigh+1
                    fprintf(fid, ['s = s + ResidualSpec.Intensity[' num2str(i) '].ToString() + "\\r\\n";\n']);
                end
                % output the scaled cross-sections:
                for j = 1:length(DOASIS.FitScenario.CrossSections)
                    fprintf(fid, ['s = s + "Fit ' DOASIS.FitScenario.CrossSections{j}.Spectrum.Name ':\\r\\n";\n']);
                    for i = DOASIS.FitScenario.ChannelRangeLow+1:DOASIS.FitScenario.ChannelRangeHigh+1
                        fprintf(fid, ['s = s + CSSpec' num2str(j) '.Intensity[' num2str(i) '].ToString() + "\\r\\n";\n']);
                    end
                end
                
                % now write everything to the output file:
                fprintf(fid, 'if (target == 1){\n');
                dout = [DOASIS.InterfaceDirectory DOASIS.OutputBuffer];
                dout = strrep(dout, '\', '\\\\');
                fprintf(fid, ['var file = fso.OpenTextFile("' dout '", 2, 1);\n']);  % for writing                          
                fprintf(fid, '}\nelse{\n');
                fprintf(fid, ['var file = fso.OpenTextFile("' dout '", 8, 0);\n']); % for appending
                fprintf(fid, '}\n');
                fprintf(fid, 'file.Write(s);\n');
                fprintf(fid, 'file.Close();\n');
            else
                for i = 1:length(DOASIS.FitScenario.CrossSections)
                    fprintf(fid, ['d = RefInfos[' num2str(i-1) '].GetCoefficientValue(0);\n']);                
                    fprintf(fid, 's = s + "<fc>" + d.ToString() + "</fc>" + "\\r\\n";\n');                
                end
                for i = 1:length(DOASIS.FitScenario.CrossSections)
                    fprintf(fid, ['d = RefInfos[' num2str(i-1) '].GetCoefficientError(0);\n']);                
                    fprintf(fid, 's = s + "<fce>" + d.ToString() + "</fce>" + "\\r\\n";\n');                
                end
                for i = 1:length(DOASIS.FitScenario.CrossSections)
                    fprintf(fid, ['d = RefInfos[' num2str(i-1) '].GetCoefficientValue(1);\n']);                
                    fprintf(fid, 's = s + "<sh>" + d.ToString() + "</sh>" + "\\r\\n";\n');                
                end
                for i = 1:length(DOASIS.FitScenario.CrossSections)
                    fprintf(fid, ['d = RefInfos[' num2str(i-1) '].GetCoefficientValue(2);\n']);                
                    fprintf(fid, 's = s + "<sq>" + d.ToString() + "</sq>" + "\\r\\n";\n');                
                end
                fprintf(fid, 's = s + "<c2>" + (fit.ChiSquare).ToString() + "</c2>" + "\\r\\n";\n');                
                                
                % now write it to the console:
                fprintf(fid, 'System.Console.WriteLine(s);\n');
            end
                                                
            %endif from dofit
            fprintf(fid, '}\n');
            
            %if the fit failed, we can throw an error here. The problem is that this will stop the script from processing any subsequent spectra. But maybe it's safer because I'm not sure what happens otherwise:
            fprintf(fid, 'else{\n');
            fprintf(fid, 'throw("The DOASIS fit threw an error.");\n}\n');
            
            % add the possibility to manually terminate the loop from DOASIS
            fprintf(fid, 'if (Script.StopAllScripts) break;\n');
            
            %end the while loop
            fprintf(fid, '}\n');                                                
            
            % close the j-script file
            fclose(fid);
        end
        
        
        function Go(DOASIS)
            % Build the jscript for execution:
            DOASIS.BuildScript([DOASIS.InterfaceDirectory DOASIS.ScriptFilename]);
                                                            
            % store the results in the private parameters of CDOASIS
            DOASIS.FitCoefficients = zeros(length(DOASIS.FitScenario.CrossSections), DOASIS.NTargets);
            DOASIS.FitCoefficientErrors = zeros(length(DOASIS.FitScenario.CrossSections), DOASIS.NTargets);
            DOASIS.Chi2 = zeros(1, DOASIS.NTargets);
            DOASIS.Shifts = zeros(length(DOASIS.FitScenario.CrossSections), DOASIS.NTargets);
            DOASIS.Squeezes = zeros(length(DOASIS.FitScenario.CrossSections), DOASIS.NTargets);
            DOASIS.FitPolynomial = zeros(DOASIS.FitScenario.PolynomialOrder+1, DOASIS.NTargets);
            DOASIS.PreFitTarget = zeros(DOASIS.FitScenario.ChannelRangeHigh - DOASIS.FitScenario.ChannelRangeLow + 1, DOASIS.NTargets);
            DOASIS.FitResidual = zeros(DOASIS.FitScenario.ChannelRangeHigh - DOASIS.FitScenario.ChannelRangeLow + 1, DOASIS.NTargets);
            DOASIS.FitCrossSections = zeros(DOASIS.FitScenario.ChannelRangeHigh - DOASIS.FitScenario.ChannelRangeLow + 1, length(DOASIS.FitScenario.CrossSections), DOASIS.NTargets);
            
            if DOASIS.QuickFit == 0
                % run the j-script!
                runjs([DOASIS.InterfaceDirectory DOASIS.ScriptFilename]);
                % now get the results...
                fid = fopen([DOASIS.InterfaceDirectory DOASIS.OutputBuffer], 'rt');  
                for t = 1:DOASIS.NTargets
                    fgetl(fid);
                    for i = 1:length(DOASIS.FitScenario.CrossSections)
                        line = fgetl(fid);
                        DOASIS.FitCoefficients(i,t) = str2double(line);                
                    end
                    fgetl(fid);
                    for i = 1:length(DOASIS.FitScenario.CrossSections)
                        line = fgetl(fid);
                        DOASIS.FitCoefficientErrors(i,t) = str2double(line);           
                    end
                    fgetl(fid);
                    line = fgetl(fid);
                    DOASIS.Chi2(t) = str2double(line);                    
                    fgetl(fid);
                    for i = 1:length(DOASIS.FitScenario.CrossSections)
                        line = fgetl(fid);
                        DOASIS.Shifts(i,t) = str2double(line);                
                    end
                    fgetl(fid);
                    for i = 1:length(DOASIS.FitScenario.CrossSections)
                        line = fgetl(fid);
                        DOASIS.Squeezes(i,t) = str2double(line);                
                    end  
                    fgetl(fid);
                    for i = 1:DOASIS.FitScenario.PolynomialOrder+1
                        line = fgetl(fid);
                        DOASIS.FitPolynomial(i,t) = str2double(line);                
                    end
                    fgetl(fid);
                    for i = 1:(DOASIS.FitScenario.ChannelRangeHigh - DOASIS.FitScenario.ChannelRangeLow + 1)
                        line = fgetl(fid);
                        DOASIS.PreFitTarget(i,t) = str2double(line);                
                    end
                    fgetl(fid);
                    for i = 1:(DOASIS.FitScenario.ChannelRangeHigh - DOASIS.FitScenario.ChannelRangeLow + 1)
                        line = fgetl(fid);
                        DOASIS.FitResidual(i,t) = str2double(line);                
                    end
                    for j = 1:length(DOASIS.FitScenario.CrossSections)
                        fgetl(fid);
                        for i = 1:(DOASIS.FitScenario.ChannelRangeHigh - DOASIS.FitScenario.ChannelRangeLow + 1)
                            line = fgetl(fid);
                            DOASIS.FitCrossSections(i,j,t) = str2double(line);                
                        end 
                    end
                end
            else % in case of QuickFit:
                % run the jscript like this:
                vals = runjs([DOASIS.InterfaceDirectory DOASIS.ScriptFilename]); 
                % now read the returned values into the CDOASIS properties:
                c = 1;
                for t = 1:DOASIS.NTargets              
                    for i = 1:length(DOASIS.FitScenario.CrossSections)                    
                        DOASIS.FitCoefficients(i,t) = vals(c);
                        c = c + 1;
                    end                    
                    for i = 1:length(DOASIS.FitScenario.CrossSections)                        
                        DOASIS.FitCoefficientErrors(i,t) = vals(c);
                        c = c + 1;                        
                    end  
                    for i = 1:length(DOASIS.FitScenario.CrossSections)                        
                        DOASIS.Shifts(i,t) = vals(c);
                        c = c + 1;                        
                    end
                    for i = 1:length(DOASIS.FitScenario.CrossSections)                        
                        DOASIS.Squeezes(i,t) = vals(c);
                        c = c + 1;                        
                    end
                    DOASIS.Chi2(t) = vals(c);
                    c = c + 1;
                end
            end
        end

        
        
        function DisplayFitResult(DOASIS)   
            if DOASIS.QuickFit == 1
                disp('Error: QuickFit is on. Spectral fit results are not available in this mode.');
            else            
                % Prepare the Figure
                frf = findobj('type','figure','name','DOASIS FitResult');
                if isempty(frf)
                    frf = figure('Name', 'DOASIS FitResult', 'NumberTitle', 'off');
                    set(frf, 'Position', [30, 50, 1000, 630]);
                    set(frf, 'color', 'w');
                else 
                    clf(frf);
                    Coord=get(frf,'Position');
                    set(frf, 'Position', [Coord(1), Coord(2), 1000, 630]);
                end               
                numplots = length(DOASIS.FitScenario.CrossSections) + 3;
                numrows = ceil(numplots/3); 

                % Now plot the target and fit result
                set(0, 'CurrentFigure', frf);
                subplot(numrows, 3, 1);                        
                plot(DOASIS.FitScenario.ChannelRangeLow:DOASIS.FitScenario.ChannelRangeHigh, exp(DOASIS.PreFitTarget(:,end)), 'b');
                hold on;            
                plot(DOASIS.FitScenario.ChannelRangeLow:DOASIS.FitScenario.ChannelRangeHigh, exp(DOASIS.PreFitTarget(:,end)-DOASIS.FitResidual(:,end)), 'r');
                xlim([DOASIS.FitScenario.ChannelRangeLow, DOASIS.FitScenario.ChannelRangeHigh]);
                
                % Get the target spectrum name:
                if isstring(DOASIS.TargetSpectrum) || ischar(DOASIS.TargetSpectrum)
                    [~,fn,~] = fileparts(DOASIS.TargetSpectrum);
                    title(fn);
                elseif iscell(DOASIS.TargetSpectrum) %TargetSpectrum is a cell array of file names
                    [~,fn,~] = fileparts(DOASIS.TargetSpectrum{end});
                    title(fn);
                else %TargetSpec is a CSpectrum object
                    if isempty(DOASIS.TargetSpectrum.SpectrumID)
                        title([FIT.TargetSpectrum.Name], 'Interpreter', 'none');
                    else
                        title([FIT.TargetSpectrum.Name ' [' num2str(FIT.TargetSpectrum.SpectrumID) ']'], 'Interpreter', 'none');
                    end                    
                end
                              
                % residual
                set(0, 'CurrentFigure', frf)
                subplot(numrows, 3, 2);
                plot(DOASIS.FitScenario.ChannelRangeLow:DOASIS.FitScenario.ChannelRangeHigh, DOASIS.FitResidual(:,end), 'b');            
                xlim([DOASIS.FitScenario.ChannelRangeLow, DOASIS.FitScenario.ChannelRangeHigh]);
                title('Fit Residual');
                % polynomial
                set(0, 'CurrentFigure', frf)
                subplot(numrows, 3, 3);
                plot(DOASIS.FitScenario.ChannelRangeLow:DOASIS.FitScenario.ChannelRangeHigh,polyval(flip(DOASIS.FitPolynomial(:,end)), DOASIS.FitScenario.ChannelRangeLow:DOASIS.FitScenario.ChannelRangeHigh), 'r');
                xlim([DOASIS.FitScenario.ChannelRangeLow, DOASIS.FitScenario.ChannelRangeHigh]);                       
                title('FitPolynomial');
                % references
                for i = 1:length(DOASIS.FitScenario.CrossSections)
                    set(0, 'CurrentFigure', frf)
                    subplot(numrows, 3, 3+i)                                
                    plot(DOASIS.FitScenario.ChannelRangeLow:DOASIS.FitScenario.ChannelRangeHigh, DOASIS.FitCrossSections(:,i,end) + DOASIS.FitResidual(:,end), 'b')
                    hold on
                    plot(DOASIS.FitScenario.ChannelRangeLow:DOASIS.FitScenario.ChannelRangeHigh, DOASIS.FitCrossSections(:,i,end), 'r')
                    xlim([DOASIS.FitScenario.ChannelRangeLow, DOASIS.FitScenario.ChannelRangeHigh]);   
                    title([DOASIS.FitScenario.CrossSections{i}.Spectrum.Name ' (S = ' num2str(DOASIS.FitCoefficients(i,end), '%.3g') ')'], 'Interpreter', 'none');
                end
                drawnow();
            end
        end
        
        function WriteResultsToFile(DOASIS, OutputFile)
            % writes the fit results to the output file. appends existing
            % files. if a new file is started, includes the header.
            [d, ~, ~] = fileparts(OutputFile);
            if ~exist(d, 'dir')
                mkdir(d);
            end 
            if ~exist(OutputFile, 'file')
                hdrstr = 'SpectrumID\t';
                for i = 1:length(DOASIS.FitScenario.CrossSections)
                    hdrstr = [hdrstr 'FitCoeff(' DOASIS.FitScenario.CrossSections{i}.Spectrum.Name ')\tFitCoeffError(' DOASIS.FitScenario.CrossSections{i}.Spectrum.Name ')\tShift(' DOASIS.FitScenario.CrossSections{i}.Spectrum.Name ')\tSqueeze(' DOASIS.FitScenario.CrossSections{i}.Spectrum.Name ')\t'];
                end
                hdrstr = [hdrstr 'FitChi2\tTimeStamp\tLatitude\tLongitude\tAltitude\tSpeed\tCourse\tGPSWarnCode\tGPSQuality\tElevationAngle\tAzimuthAngle\tExposureTime\tExposures\tMaxIntensity\tMaxIntensityFitRange\tFileName\tSpectrometerType\tSpectrometerSerialNumber\n'];                
                fid = fopen(OutputFile, 'wt');
                fprintf(fid, hdrstr);
                fclose(fid);
            end
            % if TargetSpectrum is not a CSpectrum, we need to open it to
            % retrieve its metadata:            
            if isstring(DOASIS.TargetSpectrum) || ischar(DOASIS.TargetSpectrum)
                target{1} = CSpectrum();
                target{1}.Open(DOASIS.TargetSpectrum);                
            elseif iscell(DOASIS.TargetSpectrum) %TargetSpectrum is a cell array of file names
                target = cell(length(DOASIS.TargetSpectrum),1);
                for i = 1:length(DOASIS.TargetSpectrum)
                    target{i} = CSpectrum();
                    target{i}.Open(DOASIS.TargetSpectrum{i});                    
                end                                    
            else %TargetSpec is a CSpectrum object
                target{1} = DOASIS.TargetSpectrum;
            end            
            
            fid = fopen(OutputFile, 'at'); 
            try
                for t = 1:size(target) % loop over the number of targets
                    % try to derive a valid SpectrumID if none exists
                    if isempty(target{t}.SpectrumID)
                        target{t}.SpectrumID = AssignSpectrumID(target{t}.FileName);
                    end
                    maxint = max(target{t}.Spectrum);
                    maxintfr = max(target{t}.Spectrum(DOASIS.FitScenario.ChannelRangeLow:DOASIS.FitScenario.ChannelRangeHigh));
                    fn = target{t}.FileName;
                    fn = strrep(fn, '\', '\\');
                    outstr = num2str(target{t}.SpectrumID);
                    for i = 1:length(DOASIS.FitScenario.CrossSections)
                        outstr = [outstr '\t' num2str(DOASIS.FitCoefficients(i,t)) '\t' num2str(DOASIS.FitCoefficientErrors(i,t)) '\t' num2str(DOASIS.Shifts(i,t)) '\t' num2str(DOASIS.Squeezes(i,t))];
                    end
                    outstr = [outstr '\t' num2str(DOASIS.Chi2(t)) '\t' num2str(target{t}.TimeStamp, '%.6f') '\t' num2str(target{t}.Latitude, '%.6f') '\t' num2str(target{t}.Longitude, '%.6f') '\t' num2str(target{t}.Altitude) '\t' num2str(target{t}.Speed) '\t' num2str(target{t}.Course) '\t' target{t}.GPSWarnCode '\t' num2str(target{t}.GPSQuality) '\t' num2str(target{t}.ElevationAngle) '\t' num2str(target{t}.AzimuthAngle) '\t' num2str(target{t}.ExposureTime) '\t' num2str(target{t}.Exposures) '\t' num2str(maxint) '\t' num2str(maxintfr) '\t' fn '\t' target{t}.SpectrometerType '\t' target{t}.SpectrometerSerialNumber '\n'];
                    fprintf(fid, outstr);                
                end
            catch
                msgbox('The DOASIS fit encountered an error while trying to write results to disk. This is usually due to an incorrectly formatted spectrum causing the fit to fail. The fitting may continue, but beware that results may be missing from the output files.', 'DOASIS Error');
            end
            fclose(fid);
        end
    end
end

