classdef CFitScenario < handle
    %CFITSCENARIO Fit scenario for use in the DOASISfit function.
    %   Detailed explanation goes here
    
    properties
        WavelengthRangeLow % fit range lower bound [nm]
        WavelengthRangeHigh % fit range upper bound [nm]
        CrossSections % cell array of CCrossSections 
        PolynomialOrder % order of the polynomial applied in log space during the DOASIS fit
        InterfaceDirectory % Directory which contains files needed for the DOASIS interface (e.g. empty fit scenarios) and where temporary output files will be placed.
    end
    
    properties (SetAccess = private)
        FileName % the filename where this fit scenario was last saved
        ChannelRangeLow % fit range lower bound [channel]
        ChannelRangeHigh % fit range upper bound [channel]
    end
    
    methods
        function FS = CFitScenario()
            %CFITSCENARIO Construct an instance of this class
            FS.PolynomialOrder = 3; % use 3 as a default.
            FS.InterfaceDirectory = [pwd filesep];
        end
        
        function BuildDOASISFitScenario(FS, FileName)
            for i = 1:length(FS.CrossSections)
                FS.CrossSections{i}.Spectrum.Save([FS.InterfaceDirectory 'cs' num2str(i) '.std']);
            end
            
            fid = fopen([FS.InterfaceDirectory 'bfs.js'], 'wt');
            fprintf(fid, 'import System; import DoasCore; import DoasCore.Spectra; import DoasCore.Math;\n');
            fprintf(fid, '\n');
            for i = 1:length(FS.CrossSections)
                fprintf(fid, 'var CrossSectionSpec%d : ISpectrum = Specbar.GetSpectrum("%s");\n', i, FS.CrossSections{i}.Spectrum.Name);                
            end
            
            for i = 1:length(FS.CrossSections)
                fprintf(fid, 'CrossSectionSpec%d.Open("cs%d.std");\n', i, i);                
            end  
            
            % Open an empty fit scenario with the correct number of
            % reference spectra:
            fprintf(fid, 'var fit = DoasFit.Open("emptyfs%d.fs");\n', length(FS.CrossSections));
            
            % Find a cross-section with wavelength info and use that to
            % determine the channel range for the fit
            if isempty(FS.WavelengthRangeLow) || isempty(FS.WavelengthRangeHigh)
               disp('ERROR: Wavelength fit range is not set!'); 
            end
            wli = 0;
            nchannels = length(FS.CrossSections{1}.Spectrum.Spectrum);
            for i = 1:length(FS.CrossSections)
                if (nchannels == length(FS.CrossSections{i}.Spectrum.Wavelength)) && (wli == 0)
                    wli = i;
                    FS.ChannelRangeLow = round(interp1(FS.CrossSections{i}.Spectrum.Wavelength, transpose(0:nchannels-1), FS.WavelengthRangeLow));
                    if FS.ChannelRangeLow < 0
                        FS.ChannelRangeLow = 0;
                    end
                    if FS.ChannelRangeLow > nchannels-2
                        FS.ChannelRangeLow = nchannels-2;
                    end
                    FS.ChannelRangeHigh = round(interp1(FS.CrossSections{i}.Spectrum.Wavelength, transpose(0:nchannels-1), FS.WavelengthRangeHigh));
                    if FS.ChannelRangeHigh < FS.ChannelRangeLow
                        FS.ChannelRangeHigh = FS.ChannelRangeLow+1;
                    end
                    if FS.ChannelRangeHigh > nchannels-1
                        FS.ChannelRangeHigh = nchannels-1;
                    end                    
                end                       
            end
            if wli == 0
                disp('ERROR: None of the cross-sections contain wavelength information!');
            end
            
            % Set the fit channel range                       
            fprintf(fid, 'fit.FitRanges[0].LimitLow = %d;\n', FS.ChannelRangeLow);
            fprintf(fid, 'fit.FitRanges[0].LimitHigh = %d;\n', FS.ChannelRangeHigh);
            
            % Set the polynomial order
            fprintf(fid, 'fit.PolynomialOrder = %d;\n', FS.PolynomialOrder);
            
            % Set the individual cross-sections and their fit settings
            fprintf(fid, 'var RefInfos = fit.ReferencesInfo;\n');            
            for i = 1:length(FS.CrossSections)
                fprintf(fid, 'RefInfos[%d].ReferenceSpectrum = CrossSectionSpec%d;\n', i-1, i);
                fprintf(fid, 'RefInfos[%d].ObjectKey = CrossSectionSpec%d.ObjectKey;\n', i-1, i);
                
                if strcmpi(FS.CrossSections{i}.CoefficientMode, 'fix')
                    fprintf(fid, 'RefInfos[%d].SetCoefficientMode(0, "Fixed");\n', i-1);
                    fprintf(fid, 'RefInfos[%d].SetCoefficientFixed(0, %f);\n', i-1, FS.CrossSections{i}.CoefficientFixValue);
                    fprintf(fid, 'RefInfos[%d].SetCoefficientValue(0, %f);\n', i-1, FS.CrossSections{i}.CoefficientFixValue);
                end
                
                if strcmpi(FS.CrossSections{i}.CoefficientMode, 'limit')
                    fprintf(fid, 'RefInfos[%d].SetCoefficientMode(0, "LimitHard");\n', i-1);
                    fprintf(fid, 'RefInfos[%d].SetCoefficientHighLimit(0, %f);\n', i-1, FS.CrossSections{i}.CoefficientLimitValues(2));
                    fprintf(fid, 'RefInfos[%d].SetCoefficientLowLimit(0, %f);\n', i-1, FS.CrossSections{i}.CoefficientLimitValues(1));                    
                end
                
                if strcmpi(FS.CrossSections{i}.CoefficientMode, 'link')
                    fprintf(fid, 'RefInfos[%d].SetCoefficientMode(0, "Linked");\n', i-1);
                    for j = 1:length(FS.CrossSections)
                        if strcmpi(FS.CrossSections{j}.Name, FS.CrossSections{i}.CoefficientLinkTo)
                            fprintf(fid, 'RefInfos[%d].SetCoefficientLink(0, RefInfos[%d].ObjectKey);\n', i-1, j-1);
                            break;
                        end
                        if j == length(FS.CrossSections)
                            disp(['ERROR: Can not link to unknown cross section: ' FS.CrossSections{i}.CoefficientLinkTo]);
                        end
                    end                    
                end
                
                if strcmpi(FS.CrossSections{i}.ShiftMode, 'fix')
                    fprintf(fid, 'RefInfos[%d].SetCoefficientMode(1, "Fixed");\n', i-1);
                    fprintf(fid, 'RefInfos[%d].SetCoefficientFixed(1, %f);\n', i-1, FS.CrossSections{i}.ShiftFixValue);
                    fprintf(fid, 'RefInfos[%d].SetCoefficientValue(1, %f);\n', i-1, FS.CrossSections{i}.ShiftFixValue);
                end
                
                if strcmpi(FS.CrossSections{i}.ShiftMode, 'limit')
                    fprintf(fid, 'RefInfos[%d].SetCoefficientMode(1, "LimitHard");\n', i-1);
                    fprintf(fid, 'RefInfos[%d].SetCoefficientHighLimit(1, %f);\n', i-1, FS.CrossSections{i}.ShiftLimitValues(2));
                    fprintf(fid, 'RefInfos[%d].SetCoefficientLowLimit(1, %f);\n', i-1, FS.CrossSections{i}.ShiftLimitValues(1));                    
                end
                
                if strcmpi(FS.CrossSections{i}.ShiftMode, 'link')                    
                    fprintf(fid, 'RefInfos[%d].SetCoefficientMode(1, "Linked");\n', i-1);
                    for j = 1:length(FS.CrossSections)
                        if strcmpi(FS.CrossSections{j}.Name, FS.CrossSections{i}.ShiftLinkTo)
                            fprintf(fid, 'RefInfos[%d].SetCoefficientLink(1, RefInfos[%d].ObjectKey);\n', i-1, j-1);
                            break;
                        end
                        if j == length(FS.CrossSections)
                            disp(['ERROR: Can not link to unknown cross section: ' FS.CrossSections{i}.ShiftLinkTo]);
                        end
                    end
                end
                
                if strcmpi(FS.CrossSections{i}.SqueezeMode, 'fix')
                    fprintf(fid, 'RefInfos[%d].SetCoefficientMode(2, "Fixed");\n', i-1);
                    fprintf(fid, 'RefInfos[%d].SetCoefficientFixed(2, %f);\n', i-1, FS.CrossSections{i}.SqueezeFixValue);
                    fprintf(fid, 'RefInfos[%d].SetCoefficientValue(2, %f);\n', i-1, FS.CrossSections{i}.SqueezeFixValue);
                end
                
                if strcmpi(FS.CrossSections{i}.SqueezeMode, 'limit')
                    fprintf(fid, 'RefInfos[%d].SetCoefficientMode(2, "LimitHard");\n', i-1);
                    fprintf(fid, 'RefInfos[%d].SetCoefficientHighLimit(2, %f);\n', i-1, FS.CrossSections{i}.SqueezeLimitValues(2));
                    fprintf(fid, 'RefInfos[%d].SetCoefficientLowLimit(2, %f);\n', i-1, FS.CrossSections{i}.SqueezeLimitValues(1));                    
                end
                
                if strcmpi(FS.CrossSections{i}.SqueezeMode, 'link')
                    fprintf(fid, 'RefInfos[%d].SetCoefficientMode(2, "Linked");\n', i-1);
                    for j = 1:length(FS.CrossSections)
                        if strcmpi(FS.CrossSections{j}.Name, FS.CrossSections{i}.SqueezeLinkTo)
                            fprintf(fid, 'RefInfos[%d].SetCoefficientLink(2, RefInfos[%d].ObjectKey);\n', i-1, j-1);
                            break;
                        end
                        if j == length(FS.CrossSections)
                            disp(['ERROR: Can not link to unknown cross section: ' FS.CrossSections{i}.SqueezeLinkTo]);
                        end
                    end
                end
            end
            
            % Save the fit scenario
            fn = FileName;
            fn = strrep(fn, '\', '\\');
            fprintf(fid, 'DoasFit.Save(fit, "%s");\n', fn);                       
            fclose(fid);            
            
            % now actually run the j-script:
            runjs([FS.InterfaceDirectory 'bfs.js']);
            FS.FileName = [FileName];
        end
                
    end
end

